@@ -18,6 +18,8 @@ struct hwloc_ps_process {
   char string[1024];
   char name[64];
   hwloc_bitmap_t cpuset;
+#define HWLOC_PS_ALL_UIDS ((long)-1)
+  long uid;
   int bound;
   unsigned nthreads;
   unsigned nboundthreads;
@@ -32,6 +34,7 @@ struct hwloc_ps_process {
 #define HWLOC_PS_FLAG_THREADS (1UL<<0)
 #define HWLOC_PS_FLAG_LASTCPULOCATION (1UL<<1)
 #define HWLOC_PS_FLAG_SHORTNAME (1UL<<2)
+#define HWLOC_PS_FLAG_UID (1UL<<3)
 
 int hwloc_ps_read_process(hwloc_topology_t topology, hwloc_const_bitmap_t topocpuset,
 			  struct hwloc_ps_process *proc,
@@ -40,7 +43,7 @@ int hwloc_ps_read_process(hwloc_topology_t topology, hwloc_const_bitmap_t topocp
 int hwloc_ps_foreach_process(hwloc_topology_t topology, hwloc_const_bitmap_t topocpuset,
 			     void (*callback)(hwloc_topology_t topology, struct hwloc_ps_process *proc, void *cbdata),
 			     void *cbdata,
-			     unsigned long flags, const char *only_name, const char *pidcmd);
+			     unsigned long flags, const char *only_name, long only_uid, const char *pidcmd);
 
 void hwloc_ps_free_process(struct hwloc_ps_process *proc);
 
